/*
 * MASC (Mobitex Asynchronous Communication Protocol)
 * Protocol Implementation for Linux
 * Copyright 1998 Kaz Kylheku <kaz@cafe.net>
 * This work is offered under the GNU Public License
 * as part of the Linux operating system.
 */

/*
 * $Id: mobitex_masc.c,v 1.18 1998/11/22 05:09:13 root Exp $
 */

/*
 * List of things to do:
 * - Gather and display power saving mode information in /proc/net
 * - Ignore possible stray characters between MASC frames instead of assuming
 *   that a new frame starts immediately after a CR (not likely to be needed).
 * - Interrogate the modem by sending it SENS frames as described in the
 *   definition of the MASC protocol, and perform the required reinitialization
 *   ritual if the modem does not answer.
 * - NACK an M-frame if masc_mframe_decap() fails (like if the data field
 *   contains something other than hex).
 */

#include <linux/config.h>
#include <linux/module.h>
#include <linux/if.h>		/* for IFF_* defines		*/
#include <linux/if_ether.h>	/* for struct enet_statistics	*/
#include <linux/if_arp.h>	/* for ARPHRD_* 		*/
#include <linux/skbuff.h>	/* for struct skb		*/
#include <linux/inet.h>
#include <linux/timer.h>
#include <linux/proc_fs.h>
#include <asm/byteorder.h>
#include <net/arp.h>		/* arp_find()			*/
#include <linux/ctype.h>
#include "mobitex_masc.h"

#ifdef MODULE
#define MASC_VERSION    "1.7 19981021 (MODULE)"
#else
#define	MASC_VERSION	"1.7 19981021"
#endif

#define MNAME "MOBITEX_MASC: "

/*
 * Some pre-cooked useful MASC frames, complete with checksums.
 * Why don't I use string literals here? Because these have
 * a null byte at the end which adds 1 to the size, which forces
 * you to either write the size in the array declarator, or
 * subtract 1 all over the place, two error-prone alternatives. :)
 */

static char masc_init_frame[] = {
	'^', '0', '0', '1', '0', 'B', ' ', '4', '7', 'E', ',', '0', ':', '5', 'D', '\r'
};

static char masc_getman_frame[] = {
	'^', '0', '0', '0', 'C', 'F', ' ', 'P', ':', '2', '1', '\r'
};

static char masc_getbatt_frame[] = {
	'^', '0', '0', '0', 'E', 'F', ' ', 'A', '0', '3', ':', '3', '5', '\r'
};

static char masc_getsign_frame[] = {
	'^', '0', '0', '0', 'E', 'F', ' ', 'A', '0', '2', ':', '3', '4', '\r'
};

static int masc_maxdev = MASC_NRUNIT;	/* max number of MASC devices	*/
static volatile int masc_numdev = 0;	/* counter for naming devices	*/
static int masc_proc_flag;		/* proc filesystem registration	*/

static masc_dev_t masc_list;	/* circular doubly linked list of MASCs */

static struct tty_ldisc	masc_ldisc;	/* line discipline operations	*/
static struct proc_dir_entry masc_proc;	/* proc filesystem entry	*/

/*
 * MAN numbers to send broadcasts to. We don't want to do multicasts over the
 * Mobitex network. But for now we can direct a broadcast at the station of our
 * choice using the mechanism below.  By selecting a designated host which is
 * set up with proxy ARP entries, we can use ARP to resolve IP addresses to MAN
 * numbers.  Note that this is probably a bad idea; it makes more sense for
 * mobile units to be set up with fixed arp caches. These numbers can be set
 * when loading the driver, for example: insmod mobitex_masc
 * masc_bcast=5000123,6000124 to set up masc0 to
 * send broadcasts to MAN 5000123 and masc1 to send broadcasts to 6000124.
 */

int masc_bcast[MASC_NRUNIT];

/*
 * Loopback flag used only for testing purposes; set this to non-zero at your
 * own risk.
 */

int masc_loopback;

/*
 * Allocate a new buffer for a MASC frame. Try to get this from the free list
 * maintained by the device, else allocate it dynamically. A null pointer is
 * returned on failure.
 */

struct sk_buff *masc_alloc_buf(masc_dev_t *masc, int priority)
{
	struct sk_buff *skb = __skb_dequeue(&masc->free_queue);

	if (skb == NULL) {
		skb = alloc_skb(MASC_BUF_SIZE, priority);
		skb->free = 1;	/* braindamage flag	*/
	} else {
		skb_trim(skb, 0);
	}

	return skb;
}

/*
 * Load contents of socket buffer into transmit buffer.
 */

static void masc_load_tx_buf(masc_dev_t *masc, struct sk_buff *skb)
{
	memcpy(masc->tx_buf, skb->data, skb->len);
	masc->tx_left = skb->len;
	masc->tx_head = masc->tx_buf;
}

/*
 * Scan the FH queue and clean out any old packets---ones that have not
 * received their FH acknowledgement for too long. This keeps the
 * queue from being plugged up.
 */

void masc_cleanout_old(masc_dev_t *masc)
{
	int qlen;

        qlen = skb_queue_len(&masc->fh_queue);
	while (qlen--) {
		struct sk_buff *skb = __skb_dequeue(&masc->fh_queue);
		if (jiffies - skb->when > MASC_TRANS_TIMEOUT*HZ) {
			__skb_queue_tail(&masc->free_queue, skb);
			masc->tx_dropped++;
		} else {
			__skb_queue_head(&masc->fh_queue, skb);
		}
	}
}

/*
 * Do whatever it takes to stuff the transmit buffer. Return 0 to indicate that
 * there is nothing in the transmit buffer, otherwise return 1. We don't worry
 * about reentrancy because it is assumed that the device was atomically marked
 * busy prior to calling this function---we will never be invoked by a caller
 * that finds the device to be already busy upon doing the atomic test and set.
 */

static int masc_fill_transmit_buf(masc_dev_t *masc)
{
	struct sk_buff *skb;

	/*
	 * The transmit buffer has data, so we can indicate that it's already
	 * filled. This should not happen, however, if the device just went
	 * from non-busy to busy, which is the test-and-set criterion for
	 * entering this function!  So this is more of a sanity check.
	 */

	if (masc->tx_left != 0)
		return 1;

	/*
	 * The control queue has the highest priority, so we check it first.
	 * Since control packets don't have to be retained for retransmission,
	 * once a control packet is sent, we can simply throw its buffer onto
	 * the queue of free buffers. The retransmit logic will create these
	 * packets from scratch.
	 */

	if ((skb = __skb_dequeue(&masc->ctrl_queue)) != NULL) {
		masc_load_tx_buf(masc, skb);	
		__skb_queue_head(&masc->free_queue, skb);
		return 1;
	} 

	/*
	 * Next, we check whether we are waiting for an ACK/NACK from the radio
	 * modem by checking whether the ack wait queue has a packet in it.  If
	 * we are waiting for an ACK, we cannot slide down any of the
	 * information frames, so we must sadly leave the transmit buffer
	 * empty---unless there is a request to retransmit the packet that is
	 * waiting for the ack. If we retransmit the packet, we must also
	 * enable the generation of RACKs, restart the RACK timer.
	 */

	if (skb_queue_len(&masc->ackw_queue) > 0) {
		if (clear_bit(MASC_FLG_RETRAN, &masc->flags)) {
			skb = skb_peek(&masc->ackw_queue);
			masc_load_tx_buf(masc, skb);
			del_timer(&masc->ack_timer);
			masc->ack_timer.expires = jiffies + MASC_ACK_TIMEOUT*HZ;
			add_timer(&masc->ack_timer);
			return 1;
		}
		return 0;
	}

	/*
	 * There is nothing requiring an ack, so kill the timer in
	 * case it is running---which it should not be, since either
	 * this is the very first time through this code, or we
	 * got this far because an ack was successfully received,
	 * in which case the timer would have been killed.
	 */

	del_timer(&masc->ack_timer);

	/*
	 * Clear the bit saying that we have sent a rack, since
	 * we will be starting on a brand new information frame.
	 */

	clear_bit(MASC_FLG_RACKS, &masc->flags);
	masc->ack_timer.expires = jiffies + MASC_ACK_TIMEOUT*HZ;

	/*
	 * We now check the queue of non-MPAK info frames first, giving it
	 * higher priority.  This queue could contain an INIT frame, or
	 * commands to sense the battery status, etc. None of the packets
	 * in this queue are destined for radio transmission; they are strictly
	 * between Linux and the radio modem.
	 */
	 
	if ((skb = __skb_dequeue(&masc->info_queue)) != NULL) {
		MASC_DBG_PRINTK((KERN_INFO MNAME "%s: queuing info frame\n", masc->if_name));
		__skb_queue_tail(&masc->ackw_queue, skb);
		masc_load_tx_buf(masc, skb);
		add_timer(&masc->ack_timer);
		return 1;
	}

	/*
	 * Before we check the MPAK queue, we also check whether we have have
	 * been asked to shut up at the request of the Mobitex network. If so,
	 * we politely refrain from sending any packets to the radio modem.
	 */

	if (test_bit(MASC_FLG_TSTOP, &masc->flags)) {
		MASC_DBG_PRINTK((KERN_INFO MNAME "%s: throttled by command from network\n", masc->if_name));
		return 0;
	}

	/* 
	 * Rid the FH queue of stale packets.
	 */

	masc_cleanout_old(masc);

	/*
	 * We check whether the queue of M frames which have not yet received
	 * an FH acknowledgement frame has reached the maximum. If not, we can
	 * proceed, and try to grab the next M frame from the MPAK queue and
	 * stuff it into the modem. When that packet is acked, the
	 * masc_process_ack() function will add it to the fh_queue, where it
	 * will await an FH frame from the radio modem, indicating that it has
	 * been transmitted and removed from the radio modem's own internal
	 * queue. MASC_MAX_FRAME is the maximum number of unradioed M frames
	 * that the modem can supposedly hold. If the fh_queue is currently
	 * full, we must return empty handed (or empty buffered, so to speak).
	 */	

	if (skb_queue_len(&masc->fh_queue) >= MASC_MAX_MFRAME) {
		MASC_DBG_PRINTK((KERN_INFO MNAME "%s: max outstanding M frames reached\n", masc->if_name));
		return 0;
	}

	if (masc->traf_count >= MASC_MAX_AIRED_PACKETS) {
		MASC_DBG_PRINTK((KERN_INFO MNAME "%s: max aired packets reached, backing off\n", masc->if_name));
		return 0;
	}

	/*
	 * Try to get a frame from the MPAK queue. If it is empty, clear
	 * the busy flag to indicate that the device can take more data
	 * and return.  But don't clear the busy flag if we have not yet
	 * received our local MAN---we want to make the device appear busy
	 * until we have learned of our hardware address, so that we
	 * never transmit packets with a bogus source MAN.
	 */

	if ((skb = __skb_dequeue(&masc->mpak_queue)) == NULL) {
		if (test_bit(MASC_FLG_HAVEMAN, &masc->flags)) {
			masc->dev->tbusy = 0;
			mark_bh(NET_BH);
		}
		return 0;
	}

	/* 
	 * We have an MPAK frame to send, so we load it into the transmit
	 * buffer and put it into the one-element ACK waiting queue,
	 * then start the RACK timer.
	 */

	MASC_DBG_PRINTK((KERN_INFO MNAME "%s: sending M frame\n", masc->if_name));
	__skb_queue_tail(&masc->ackw_queue, skb);
	masc_load_tx_buf(masc, skb);
	add_timer(&masc->ack_timer);
	return 1;
}

/*
 * Send data down to tty to initiate data flow, if it is not transmitting
 * already! This should be called whenever data has been made available on the
 * queues.
 */

void masc_kick_tty(masc_dev_t *masc)
{
	struct device *dev = masc->dev;

	/*
	 * We check if the device is already busy transmitting data.  If it is
	 * busy, and the transmit period has not expired, we simply return. If
	 * the transmit period *has* expired, something went wrong with the
	 * transmit, and so we simply clear it and start a new one. We
	 * therefore have lost a packet, so we increment the error count.
	 */

	if (set_bit(MASC_FLG_TBUSY, &masc->flags)) {
		if (jiffies - dev->trans_start < 20 * HZ) {
			MASC_DBG_PRINTK((KERN_INFO MNAME "device %s busy transmitting\n", masc->if_name));
			return;
		}
		dev->trans_start = jiffies;
		printk(KERN_ERR MNAME "transmit timed out for %s\n", masc->if_name);
		masc->tx_errors++;
		masc->tx_left = 0;
	}

	/*
	 * We have now marked the device as busy, and so we test whether we can
	 * deliver data.  We might not be able to, even though there are queued
	 * frames, because we are waiting for an ack, or are out of contact
	 * with the network, etc.  If we can't deliver data, we must clear the
	 * busy flag and disable write wakeup callbacks.  Otherwise, we enable
	 * write wakeups and stuff the data into the tty device to kick-start
	 * the downward flow.
	 */

	if (!masc_fill_transmit_buf(masc)) {
		masc->tty->flags &= ~(1 << TTY_DO_WRITE_WAKEUP);
		clear_bit(MASC_FLG_TBUSY, &masc->flags);
	} else {
		int nwrit;
		struct tty_struct *tty = masc->tty;
		masc->dev->trans_start = jiffies;
		tty->flags |= (1 << TTY_DO_WRITE_WAKEUP);
		nwrit = tty->driver.write(tty, 0, masc->tx_head, masc->tx_left);
		masc->tx_left -= nwrit;
		masc->tx_head += nwrit;
	}
}

/*
 * Read the specified number of hexadecimal characters from the given buffer,
 * and convert to a long value.
 */

static long masc_read_hex(unsigned char *buf, int len)
{
	long hex = 0;

	while (len--) {
		int ch = *buf++;
		if (isdigit(ch)) {
			ch -= '0';
		} else {
			char *h = "ABCDEF";
			char *p = strchr(h, ch);
			if (!p)
				return -1;
			ch = p - h + 10;
		}
		hex <<= 4;
		hex |= ch;
	}

	return hex;
}


/*
 * Check an incoming MASC information or init packet for basic criteria:
 * - starts with ^ character, indicating it's an info frame
 *   (assumed to be checked by the caller).
 * - contains exactly one colon (:) character;
 * - has a total length of at least 10 characters;
 * - ends with a carriage return (assumed checked by the caller).
 * An information packet which does not meet these criteria is
 * discarded without. An information packet which meets these
 * criteria but does not meet the full criteria requires
 * a NACK response, according to the MASC protocol.
 */

static int masc_infoframe_basic_validate(unsigned char *buf, int len)
{
	unsigned char *p;

	if (len < 10)
		return 0;

	if (!(p = memchr(buf, ':', len)))
		return 0;

	if (memchr(p + 1, ':', len - (p + 1 - buf)))
		return 0;

	return 1;
}

/*
 * Check an incoming MASC information or init packet for the full required
 * criteria:
 * - starts with ^ character;
 * - contains exactly one colon (:) character;
 * - ``check'' and ``length'' fields have correct values;
 * - no character with parity error exists;
 * - maximum number of characters has not been exceeded in any
 *   field or frame;
 * - ends with carriage return.
 * It is assumed that the packet has already passed the basic
 * criteria checked by masc_infoframe_basic_validate, and that the
 * parity errors are checked elsewhere. So we just verify the
 * checksums and sizes.
 */

static int masc_infoframe_validate(unsigned char *buf, int len)
{
	long length;
	int cksum = 0;

	length = masc_read_hex(buf + 1, 4);

	if (length != len) {
		MASC_DBG_PRINTK((KERN_INFO MNAME "validation failed due to bad length\n"));
		return 0;
	}

	while (len-- > 3)
		cksum ^= *buf++;

	MASC_DBG_PRINTK((KERN_INFO MNAME "computed cksum is %x\n", (unsigned) cksum));
	MASC_DBG_PRINTK((KERN_INFO MNAME "scanned cksum is %x\n", (unsigned) masc_read_hex(buf, 2)));
	return cksum == masc_read_hex(buf, 2);
}

/*
 * Send a generic control frame. The given string provides the one
 * or two character packet contents. Since the contents are variable,
 * we must compute the checksum field. It is required that
 * strlen(pkt) <= 2.
 */

static void masc_send_ctrl(masc_dev_t *masc, char *pkt)
{
	struct sk_buff *skb = masc_alloc_buf(masc, GFP_ATOMIC);
	unsigned char *out;

	if (!skb) {
		printk(KERN_ERR MNAME "%s: unable to send control frame due to insufficient memory!\n", masc->if_name);
		return;
	}

	out = skb->data;

	*out++ = '^';
	(void) ((*out++ = *pkt++) && (*out++ = *pkt++));
	*out++ = '\r';

	skb_put(skb, out - skb->data);
	skb->protocol = MASC_CTRL;
	__skb_queue_tail(&masc->ctrl_queue, skb);
}

/*
 * Send the given ack. The ack parameter must be -1, 0, or 1,
 * and gets converted to the characters '-', '0' or '1'. An
 * ack of '-' means that we have been asked to re-ack (via a RACK
 * frame) but we have never sent an ACK before.
 */

static void masc_send_ack(masc_dev_t *masc, int ack)
{
	char ackstr[3];

	ackstr[0] = '*';
	ackstr[1] = ack["-01"+1];
	ackstr[2] = 0;

	masc_send_ctrl(masc, ackstr);
}

/*
 * Process an ACK frame received from the modem, according to
 * the rules of the MASC protocol.
 */

static void masc_process_ack(masc_dev_t *masc, unsigned char *pak, int len)
{
	int repeat_frame = 0;

	if (len != 4)
		return;

	if (clear_bit(MASC_FLG_RACKS, &masc->flags)) {
		switch (pak[2]) {
		case '0':
			if (masc->last_ack != 1)
				repeat_frame = 1;
			masc->last_ack = 0;
			break;
		case '1':
			if (masc->last_ack != 0)
				repeat_frame = 1;
			masc->last_ack = 1;
			break;
		case '-':
			repeat_frame = 1;
			break;	
		default:
			break;
		}
	} else {
		if (pak[2] == '0' || pak[2] == '1')
			masc->last_ack = pak[2] - '0';
	}

	if (repeat_frame) {
		set_bit(MASC_FLG_RETRAN, &masc->flags);
		MASC_DBG_PRINTK((KERN_INFO MNAME "ack resulting in retransmit\n"));
	} else {
		struct sk_buff *skb;

		/*
		 * atomically move ack packet to either free_queue or fh_queue,
		 * disable the ack timer and clear the retransmit flag.
		 */

		skb = __skb_dequeue(&masc->ackw_queue);
		if (skb) {
			if (skb->protocol == MASC_INFO) {
				__skb_queue_head(&masc->free_queue, skb);
			} else {
				/*
				 * timestamp used for cleaning out stale entries
				 * from FH queue
				 */
				skb->when = jiffies;
				__skb_queue_tail(&masc->fh_queue, skb); 
			}
		} else {
			printk(KERN_ERR MNAME "%s: got ack, but ack queue empty!\n", masc->if_name);
		}
		clear_bit(MASC_FLG_RETRAN, &masc->flags);
		del_timer(&masc->ack_timer);

		MASC_DBG_PRINTK((KERN_INFO MNAME "ack processed\n"));
	}

	masc_kick_tty(masc);
}

/*
 * Process a NACK frame from the modem by retransmitting the
 * current ``ack packet''. We don't actually retransmit here,
 * but we set a flag indicating that a retransmit is needed.
 * This is detected in masc_fill_transmit_buf which is called
 * by masc_kick_tty.
 */

static void masc_process_nack(masc_dev_t *masc, unsigned char *pak, int len)
{
	if (len != 3)
		return;

	clear_bit(MASC_FLG_RACKS, &masc->flags);
	set_bit(MASC_FLG_RETRAN, &masc->flags);
	MASC_DBG_PRINTK((KERN_INFO MNAME "nack resulting in retransmit\n"));
	masc_kick_tty(masc);
}

/*
 * Search the queue of outstanding MASC packets (ones that not been sent over
 * the radio yet) and find one that matches the given sequence number,
 * otherwise return null. We disable interrupts while searching the
 * queue, and use unsafe operations. The queue is never longer than
 * MASC_MAX_FRAME, which is a small number---typically 2.
 * Instead of dequeuing packets and placing them to the
 * rear of the queue, how about looping through skb->next here? 
 * This would, however, violate the opaque sk_buff datatype,
 * and the kernel doesn't provide a skb_next() operation.
 */

struct sk_buff *masc_get_untransmitted(masc_dev_t *masc, int seqno)
{
	int qlen;

        qlen = skb_queue_len(&masc->fh_queue);
	while (qlen--) {
		struct sk_buff *skb = __skb_dequeue(&masc->fh_queue);
		if (skb->data[7] == seqno)
			return skb;
		__skb_queue_tail(&masc->fh_queue, skb);
	}

	return NULL;
}

/*
 * Send an info frame down the info_queue. A complete MASC frame
 * with correct checksum must be supplied. M frames should not
 * be sent this way; they belong to the mpak_queue.
 */

static void masc_send_info(masc_dev_t *masc, char *frame, int len)
{
	struct sk_buff *skb = masc_alloc_buf(masc, GFP_ATOMIC);

	if (!skb) {
		printk(KERN_ERR MNAME "%s: unable to send info frame due to insufficient memory!\n", masc->if_name);
		return;
	}

	memcpy(skb->data, frame, len);
	skb_put(skb, len);
	skb->protocol = MASC_INFO;
	__skb_queue_tail(&masc->info_queue, skb);
}

/*
 * Allocate a sk_buff and decapsulate the given MASC M frame into
 * it.
 */

static struct sk_buff *masc_mframe_decap(masc_dev_t *masc, unsigned char *pak, int len)
{
	struct sk_buff *skb = masc_alloc_buf(masc, GFP_ATOMIC);
	int count = 0;
	unsigned char *out;

	if (!skb)
		return NULL;

	pak += 7;	/* skip to the MASC data field  ^0000M:<data> 	*/
	len -= 7;	/*                                ^ seven chars	*/

	out = skb->data;

	/*
	 * Extract pairs of hex digits, but ignore the last
	 * three characters (the two digit checksum and carriage return).
	 * Convert the pairs into bytes. If an invalid hex digit
	 * is detected, free the buffer and bail out.
	 */

	while (len > 3) {
		int ch = masc_read_hex(pak, 2);
		if (ch == -1) {
			__skb_queue_head(&masc->free_queue, skb);
			return NULL;
		}
		*out++ = ch;
		count++;
		pak += 2;
		len -= 2;
	}

	skb_put(skb, count);
	return skb;
}

/*
 * Process received M frame. If it is a HPDATA type M frame whose protocol
 * ID (HPID) matches one of our three supported protocol types, we pass
 * it up to the networking layer. Also, if a packet is returned by the
 * network, we simply drop it.
 */

static void masc_process_mframe(masc_dev_t *masc, unsigned char *pak, int len)
{
	int tfc_state, class, type;
	unsigned char *mpak;
	struct sk_buff *skb;
	
	if ((skb = masc_mframe_decap(masc, pak, len)) == NULL) {
		printk(KERN_ERR MNAME "%s: dropping received packet due to insufficient memory\n", masc->if_name);
		masc->rx_dropped++;
		return;
	}

	mpak = skb->data;

	tfc_state = (mpak[6] & MBTX_TFC_MASK) >> MBTX_TFC_SHIFT;
	class = (mpak[7] & MBTX_CLASS_MASK) >> MBTX_CLASS_SHIFT;
	type = (mpak[7] & MBTX_TYPE_MASK) >> MBTX_TYPE_SHIFT;

	/*
	 * The default protocol is ETH_P_MOBITEX, which marks a Mobitex packet
	 * that we have not otherwise recognized as an IP, ARP or RARP packet.
	 * Applications can install taps for this packet type to do raw Mobitex
	 * communications.
	 */

	skb->protocol = htons(ETH_P_MOBITEX);
	skb->dev = masc->dev;
	skb->mac.raw = skb->data;

	switch (tfc_state) {
	case MBTX_FROM_MAIL:
		masc->rx_mailbox++;
		/* fall through */
	case MBTX_OK:
		switch (class) {
		case MBTX_PSUBCOM:
			switch (type) {
			case MBTX_HPDATA:
				switch (mpak[11]) {
				case MBTX_P_IP:
					skb->protocol = htons(ETH_P_IP);
					break;
				case MBTX_P_ARP:
					skb->protocol = htons(ETH_P_ARP);
					break;
				case MBTX_P_RARP:
					skb->protocol = htons(ETH_P_RARP);
					break;
				default:
					break;
				}
				/*
				 * Rip off the Mobitex HPDATA header. But for packets that are
				 * not HPDATA, we don't do this. They can only be seen by taps
				 * anyway.
				 */
				skb_pull(skb, MBTX_HEADER_LEN); 
				break;
			default:
				break;
			}
			break;
		case MBTX_DTESERV:
			switch (type) {
			case MBTX_GROUPLIST:
				/*
				 * We have received a group list. If there are
				 * at leat two addresses in it, take the second
				 * one and install it as a hardware broadcast
				 * address. The first address is the ``All
				 * Terminals Group'' MAN, which cannot be used
				 * for broadcasts in most Mobitex networks. The
				 * second address is most likely a group
				 * representing all of the modems belonging to
				 * a particular subscriber, so this is a more
				 * appropriate choice.  Of course, if the user
				 * has overriden the broadcast address using
				 * the masc_bcast[] array, then don't install
				 * the modem's address.
				 */
				MASC_DBG_PRINTK((KERN_INFO MNAME "group list received\n"));
				if (mpak[8] > 1 && masc_bcast[masc->line] == 0) {
					struct device *dev = masc->dev;
					dev->broadcast[0] = mpak[12];
					dev->broadcast[1] = mpak[13];
					dev->broadcast[2] = mpak[14];
				}
				break;
			default:
				break;
			}
			break;
		default:
			break;
		}
		break;
	case MBTX_IN_MAIL:
		masc->tx_mailbox++;
		break;
	case MBTX_NO_TRANSFER:
		masc->rx_notran++;
		break;
	case MBTX_ILLEGAL:
		masc->rx_illegal++;
		break;
	case MBTX_CONGEST:
		masc->rx_congest++;
		break;
	case MBTX_ERROR:
		masc->rx_errcount++;
		break;
	case MBTX_BUSY:
		masc->rx_busy++;
		break;
	}


	MASC_DBG_PRINTK((KERN_INFO MNAME "MPAK protocol is %x\n", (unsigned) ntohs(skb->protocol)));
	netif_rx(skb);
	masc->rx_packets++;
	return;
}

/*
 * Handle error code from a F K, R or N frame. The packet sequence number is
 * specified as -1 if the error does not correspond to any packet.
 */

static void masc_handle_error(masc_dev_t *masc, int code, int seq)
{
	struct sk_buff *skb = NULL;
	enum { kick, drop, requeue } action = kick;

	if (seq != -1)
		skb = masc_get_untransmitted(masc, seq);

	if (code != -1) {
		masc->errhist[0] = masc->errhist[1];
		masc->errhist[1] = masc->errhist[2];
		masc->errhist[2] = masc->errhist[3];
		masc->errhist[3] = code;
	}

	switch (code) {
	case -1:
		action = drop;
		break;
	case 0x01:	/* modem received order to stop traffic		*/
		set_bit(MASC_FLG_TSTOP, &masc->flags);
		break;
	case 0x02:	/* modem received order to start traffic	*/
		clear_bit(MASC_FLG_TSTOP, &masc->flags);
		action = kick;
		break;
	case 0x10:	/* MPAK returned due to network stop order	*/
		set_bit(MASC_FLG_TSTOP, &masc->flags);
		action = requeue;
		break;
	case 0x13:	/* MPAK returned---transmission buffer full	*/
		printk(KERN_ERR MNAME "%s radio modem bounced packet due to buffer full!\n",
			masc->if_name);
		action = requeue;
		break;
	case 0x18:	/* The originating MAN is unrecognized		*/
		printk(KERN_ERR MNAME "%s radio modem bounced packet due to wrong originating address!\n",
			masc->if_name);
		action = drop;
		break;
	case 0x15:	/* Invalid MPAK type, must not be transmitted	*/
		action = drop;
		break;
	case 0x52:	/* Invalid MPAK state				*/
	case 0x53:	/* Invalid MPAK flags				*/
	case 0x54:	/* Invalid send list				*/
	case 0x55:	/* Invalid MPAK length				*/
	case 0x56:	/* Invalid addressee in MPAK			*/
	case 0x62:	/* MPAK returned on request (F I or F O)	*/
		action = drop;
		break;
	case 0xA0:	/* Transmission buffer full			*/
	case 0xE0:	/* Radio transmission has failed		*/
	case 0xF0:	/* Transmit queue full; max number of MPAKS	*/
	case 0xFD:	/* Other error; modem not able to transmit	*/
		action = requeue;
		break;

	/*
	 * These errors we either don't expect, cannot do anything
	 * about, or they are just informative messages that require
	 * no action.
	 */
	case 0x0A:	/* reception buffer full			*/
	case 0x0B:	/* reception buffers free			*/
	case 0x16:	/* Login denied, MAN already in flexlist	*/
	case 0x61:	/* TEMP_DEFAULT_LIST received incorrectly	*/
	case 0x63:	/* User changes state into already active mode	*/
	case 0x65:	/* Mode change already activated		*/
	case 0x66:	/* Invalid network ID				*/
	case 0x67:	/* Mode MPAK with invalid parameter		*/
	default:	/* unknown or garbled */
		action = drop;	/* just in case, so we can keep going	*/
		break;
	}

	if (skb) {
		switch (action) {
		case drop:
			__skb_queue_head(&masc->free_queue, skb);
			masc->tx_dropped++;
			break;
		case requeue:
			__skb_queue_tail(&masc->mpak_queue, skb);
			masc->tx_errors++;
			break;
		case kick:
			__skb_queue_tail(&masc->fh_queue, skb);
			break;
		}
		masc_kick_tty(masc);
	}
}

/*
 * Process received information frame based the type specified in its
 * text field. A wide variety of information is communicated through
 * info frames, in addition to MPAKS. We perform additional sanity
 * checking on the information frames as opportunity arises, and
 * return 0 so that the caller can NACK a badly received packet.
 */

static int masc_process_infoframe(masc_dev_t *masc, unsigned char *pak, int len)
{
	int seqno, err, value;
	long man;
	struct sk_buff *skb;

	switch (pak[5]) {
	case 'B':		/* INIT frame		*/
		/*
		 * Send INIT frame to the modem. Should we really do this now?
		 * It's not clear whether the protocol requires it but it works.
		 */
		masc_send_info(masc, masc_init_frame, sizeof masc_init_frame);
		/*
		 * Send request to obtain modem's MAN address if we don't
		 * already have it.
		 */
		if (!test_bit(MASC_FLG_HAVEMAN, &masc->flags))
			masc_send_info(masc, masc_getman_frame, sizeof masc_getman_frame);
		break;
	case 'M':		/* MPAK-carrying frame	*/
		masc_process_mframe(masc, pak, len);
		break;
	case 'F':
		if (pak[6] != ' ')
			return 0;
		switch (pak[7]) {
		case 'H':	/* FH frame---MPAK radioed to base	*/
			MASC_DBG_PRINTK((KERN_INFO MNAME "FH frame (%c) being processed\n", (int) pak[8]));
			seqno = pak[8];
			if (!isdigit(seqno))
				return 0;
			skb = masc_get_untransmitted(masc, seqno);
			if (skb) {
				__skb_queue_head(&masc->free_queue, skb);
				masc->tx_packets++;
			} else {
				printk(KERN_INFO "%s: received FH frame for packet it didn't send\n",
					masc->if_name);
			}
			masc->traf_hist[(masc->traf_curr + MASC_MAX_AVG_AIRTIME) % (sizeof masc->traf_hist / sizeof masc->traf_hist[0])]++;
			masc->traf_count++;
			/* if we are getting FH's we must be in coverage */
			clear_bit(MASC_FLG_NOCONTACT, &masc->flags);
			break;
		case 'K':	/* FK frame---error condition		*/
			err = masc_read_hex(pak + 8, 2);
			masc_handle_error(masc, err, -1);
			break;	
		case 'G':	/* Lost contact with Mobitex network	*/
			set_bit(MASC_FLG_NOCONTACT, &masc->flags);
			masc_kick_tty(masc);
			break;
		case 'F':	/* In contact with Mobitex network 	*/
			clear_bit(MASC_FLG_NOCONTACT, &masc->flags);
			break;
		case 'P':	/* Modem is telling us the MAN number!	*/
			if (len != 18)
				return 0;
			man = masc_read_hex(pak + 8, 6);	
			if (man != -1 && man != 0 && man != 1) {
				unsigned char *addr = masc->dev->dev_addr;
				*addr++ = (man >> 16) & 0xff;
				*addr++ = (man >> 8) & 0xff;
				*addr++ = man & 0xff;
				set_bit(MASC_FLG_HAVEMAN, &masc->flags);
			}
			break;
		case 'A':	/* Battery, RSSI or power save info	*/
			if (len != 17)
				return 0;
			if (pak[10] != ',')
				return 0;
			switch (masc_read_hex(pak + 8, 2)) {
			case -1:	/* bad hex */
				return 0;
			case 1:	/* power saving mode info	*/
				break;	/* not supported yet	*/
			case 2:	/* RSSI value			*/
				value = masc_read_hex(pak + 11, 2);
				if (value == -1)
					return 0;
				masc->sign = value;
				set_bit(MASC_FLG_HAVERSSI, &masc->flags);
				break;
			case 3: /* Battery value		*/
				value = masc_read_hex(pak + 11, 2);
				if (value == -1)
					return 0;
				masc->batt = value;
				set_bit(MASC_FLG_HAVEBATT, &masc->flags);
				break;
			}
			break;
		default:	/* Unsupported F frame			*/
			break;
		}
		break;
	case 'N':		/* Return of untransmitted MPAK		*/
	case 'R':		/* Return of errnoeous MPAK		*/
		if (pak[6] == '\r')
			return 1;
		if (pak[6] != ' ')
			return 0;
		if (pak[7] == ',') {
			err = -1;
			seqno = pak[8];
		} else {
			if (pak[9] != ',')
				return 0;
			err = masc_read_hex(pak + 7, 2);
			seqno = pak[10];
		}
		masc_handle_error(masc, err, seqno);
		break;	
	default:		/* drop unrecognized info frame		*/
		break;
	}
	return 1;
}

/*
 * Process frame received from modem as a MASC packet, taking
 * into account the overflow and toss (parity error) flag.
 */

static void masc_process_frame(masc_dev_t *masc)
{
	unsigned char *rx_buf = masc->rx_buf;
	int rx_count = masc->rx_count;
	int overflow = clear_bit(MASC_FLG_OFLOW, &masc->flags);
	int toss = clear_bit(MASC_FLG_TOSS, &masc->flags);

	MASC_DBG_PRINTK((KERN_INFO MNAME "processing received frame, length %x\n", (unsigned) rx_count));
	rx_buf[rx_count] = 0;
	MASC_DBG_PRINTK((KERN_INFO MNAME "frame (truncated to 64 chars): %.64s\n", rx_buf));
	masc->rx_count = 0;

	if (overflow)
		return;

	if (rx_count < 2)
		return;

	if (rx_buf[0] != '^' && rx_buf[rx_count-1] != '\r')
		return;

	if (isxdigit(rx_buf[1])) {	/* information or init frame	*/
		if (masc_infoframe_basic_validate(rx_buf, rx_count)) {
			if (!toss && masc_infoframe_validate(rx_buf, rx_count)
					&& masc_process_infoframe(masc, rx_buf, rx_count)) {
				masc->send_ack += 1;
				masc->send_ack &= 1;
				masc_send_ack(masc, masc->send_ack); 
			} else {
				masc_send_ctrl(masc, "?");	/* send NACK */
			}
			masc_kick_tty(masc);
		}
	} else {			/* control or garbled frame	*/
		if (toss)
			return;
		switch (rx_count) {
		case 4:
			if (rx_buf[1] == '*')	/* ACK frame */
				masc_process_ack(masc, rx_buf, rx_count);
			break;
		case 3:
			switch (rx_buf[1]) {
			case '?':		/* NACK frame	*/
				masc_process_nack(masc, rx_buf, rx_count);
				break;
			case '!':		/* RACK frame	*/
				masc_send_ack(masc, masc->send_ack);
				masc_kick_tty(masc);
				break;
			case '#':		/* SENS frame	*/
				masc_send_ctrl(masc, "&");
				masc_kick_tty(masc);
				break;
			case '&':		/* SACK frame	*/
				break;
			default:		/* whatever	*/
				break;
			}
		default:
			break;
		}
	}
}

/*
 * Free the buffers on all the MASC queues.
 * This is performed when the device is shut down.
 * At this point, all timers and other bottom-half callbacks
 * have stopped and we are disconnected from the networking layer.
 * So it's safe to use non-atomic queue operations.
 */

static void masc_free_buffers(masc_dev_t *masc)
{
	struct sk_buff *skb;

	while ((skb = __skb_dequeue(&masc->mpak_queue)) != NULL)
		kfree_skb(skb, FREE_WRITE);

	while ((skb = __skb_dequeue(&masc->info_queue)) != NULL)
		kfree_skb(skb, FREE_WRITE);

	while ((skb = __skb_dequeue(&masc->ctrl_queue)) != NULL)
		kfree_skb(skb, FREE_WRITE);

	while ((skb = __skb_dequeue(&masc->free_queue)) != NULL)
		kfree_skb(skb, FREE_WRITE);

	while ((skb = __skb_dequeue(&masc->ackw_queue)) != NULL)
		kfree_skb(skb, FREE_WRITE);

	while ((skb = __skb_dequeue(&masc->fh_queue)) != NULL)
		kfree_skb(skb, FREE_WRITE);
}

/*
 * Network device operations.
 *
 */

/*
 * Function called by the networking layer to indicate that the network
 * device is going up! We don't have to do anything here, but we perform
 * a simple sanity check to ensure we are connected to a tty.
 */

static int masc_dev_open(struct device *dev)
{
	masc_dev_t *masc;
	struct tty_struct *tty;

	if (!(masc = dev->priv)) {
		printk(KERN_ERR MNAME "null pointer to masc structure in net device!\n");
		return -ENXIO;
	}

	if (!(tty = masc->tty)) {
		printk(KERN_ERR MNAME "interface wants to go up, but tty is unavailable!\n");
		return -ENXIO;
	}

	return 0;
}

static int masc_dev_ioctl(struct device *dev, struct ifreq *ifr, int cmd)
{
	return -EINVAL;	/* no ioctl's for now */
}

/*
 * Stop the device from transmitting
 */

static int masc_dev_stop(struct device *dev)
{
	masc_dev_t *masc = dev->priv;

	if (masc->tty == NULL) {
		return -EBUSY;
	}

	/*
	 * Make it appear as though the device is in a
	 * busy state, and prevent tty wakeups.
	 */

	masc->tty->flags &= ~(1 << TTY_DO_WRITE_WAKEUP);
	dev->tbusy = 1;

	return 0;
}

/*
 * Take a raw packet and encapsulate it as a MASC info frame.
 */


static void masc_encaps_mframe(masc_dev_t *masc, struct sk_buff *sk_in, struct sk_buff *sk_out)
{
	unsigned char *in = sk_in->data, *out = sk_out->data;
	int len = 1 + 4 + 3 + 1 + sk_in->len * 2 + 2 + 1;
	int cksum, count = sk_in->len;

	cksum = *out++ = '^';

	sprintf(out, "%04X", len);
	cksum ^= *out++;
	cksum ^= *out++;
	cksum ^= *out++;
	cksum ^= *out++;

	cksum ^= *out++ = 'M';
	cksum ^= *out++ = ' ';
	cksum ^= *out++ = '0' + masc->seqno;
	cksum ^= *out++ = ':';

	while (count--) {
		cksum ^= *out++ = ((*in >> 4) & 0xf)["0123456789ABCDEF"];
		cksum ^= *out++ = (*in++ & 0xf)["0123456789ABCDEF"];
	}

	*out++ = ((cksum >> 4) & 0xf)["0123456789ABCDEF"];
	*out++ = (cksum & 0xf)["0123456789ABCDEF"];

	*out++ = '\r';

	skb_put(sk_out, out - sk_out->data);

	if (masc->seqno++ == 9)
		masc->seqno = 0;

	sk_out->protocol = MASC_MPAK;
}

/*
 * Called from the networking system to pass a packet down to us, which
 * we turn into an M frame and ship it to the modem.
 */

static int masc_dev_xmit(struct sk_buff *skb, struct device *dev)
{
	masc_dev_t *masc = dev->priv;
	struct sk_buff *masc_skb;

	MASC_DBG_PRINTK((KERN_INFO MNAME "masc_dev_xmit on %s\n", masc->if_name));

	/*
	 * If this is an IP packet, we see whether the queue already has an IP
	 * packet that is going to the same Mobitex destination, and which has
	 * enough room to accomodate this IP packet. The primary reason for this
	 * is to aggregate non-piggybacked TCP acknowledgements.
	 */

	/*
	 * We check the size of our M frame queue. If it is too long,
	 * we tell the networking system to hold its horses.
	 */

	if (skb_queue_len(&masc->mpak_queue) >= MASC_MAX_QUEUE) {
		MASC_DBG_PRINTK((KERN_INFO MNAME "masc_dev_xmit on %s balked: too many packets queued\n", masc->if_name));
		dev->tbusy = 1;		/* indicate that device can't take any more packets */
		masc_kick_tty(masc);	/* kick-start downward flow in case it's stopped */
		return 1;
	}

	/*
	 * Otherwise we try to allocate a new MASC packet buffer.
	 * If we can't allocate packet, we simply balk.
	 */

	if ((masc_skb = masc_alloc_buf(masc, GFP_ATOMIC)) == NULL) {
		printk(KERN_ERR MNAME "%s: masc_dev_xmit dropping packet due to insufficient memory\n", masc->if_name);
		dev_kfree_skb(skb, FREE_WRITE);	/* just toss the packet */
		return 0;
	}

	/*
	 * We take the raw MPAK and encapsulate it into an M frame MASC packet,
	 * giving an appropriate sequence number, etc.
	 * After this point, we have accepted the packet, and so it is
	 * our responsibility to free the sk_buff we have been given
	 * and to return a zero.  The header construction has already been done
	 * for this packet so it is a valid Mobitex PSUBCOM HPDATA packet.
	 */

	masc_encaps_mframe(masc, skb, masc_skb);
	dev_kfree_skb(skb, FREE_WRITE);

	MASC_DBG_PRINTK((KERN_INFO MNAME "%s: encapsulated frame\n", masc->if_name));

	/*
	 * We stuff the packet down into the queue designated for MPAK-carrying
	 * MASC frames (M frames).
	 */

	__skb_queue_tail(&masc->mpak_queue, masc_skb);

	/*
	 * We kick some data down to the tty if it is not already
	 * busy transmitting.
	 */

	masc_kick_tty(masc);
	return 0;
}

/*
 * Gather traffic statistics into a standard form. One of the places
 * this information ends up is on the output of the ifconfig command,
 * allowing the user to view the statistics.
 */

static struct enet_statistics *masc_dev_stats(struct device *dev)
{
	static struct enet_statistics stats;
	masc_dev_t *masc = dev->priv;

	memset(&stats, 0, sizeof stats);

	stats.rx_packets = masc->rx_packets;
	stats.tx_packets = masc->tx_packets;
	stats.tx_errors = masc->tx_errors;
	stats.rx_errors = masc->rx_errors;
	stats.tx_dropped = masc->tx_dropped;
	stats.rx_dropped = masc->rx_dropped;

	return &stats;
}

/*
 * We are called upon by the network layer to build a Mobitex header.
 */

static int masc_dev_hard_header(struct sk_buff *skb, struct device *dev,
		unsigned short type, void *daddr,
		void *saddr, unsigned int len)
{
	int hpdata;
	unsigned char *hdr = skb_push(skb, MBTX_HEADER_LEN);

	MASC_DBG_PRINTK((KERN_INFO MNAME "masc_dev_hard_header called\n"));

	/*
	 * These are the packet types we support, and map to MOBITEX HPDATA
	 * packets under various HPID's. This is a terrible hack, but what can
	 * we do? The network layers think we are an ethernet device, so they
	 * give us raw ethernet enumerations of packet types.
	 */

	switch (type) {
	case ETH_P_IP:
		hpdata = MBTX_P_IP;
		break;
	case ETH_P_ARP:
		hpdata = MBTX_P_ARP;
		break;
	case ETH_P_RARP:
		hpdata = MBTX_P_RARP;
		break;
	default:
		printk(KERN_ERR MNAME "%s won't do %u protocol\n", dev->name, type);
		hpdata = MBTX_P_BAD;	/* sentinel marking invalid protocol */
		break;
	}

	/*
	 * If we are given a source address, then use it, otherwise
	 * use our own.
	 */

        if (saddr)
                memcpy(hdr, saddr,dev->addr_len);
        else
                memcpy(hdr, dev->dev_addr, dev->addr_len);

	/*
	 * Set the Mobitex subscription flags and traffic state 
	 * byte to the appropriate values.
	 */

	hdr[6] = MBTX_OUTGOING_FLAGS;
	hdr[7] = MBTX_OUTGOING_TYPE;
	hdr[8] = 0;	/* time field, we just zero it out */
	hdr[9] = 0;
	hdr[10] = 0;
	hdr[11] = hpdata;

	/*
 	 * Set the destination address if it is available,
	 * otherwise return the negated header length
	 */

        if (daddr) {
                memcpy(hdr + 3, daddr, dev->addr_len);
                return MBTX_HEADER_LEN;
        } else {
		if (masc_loopback)
			memcpy(hdr + 3, dev->dev_addr, dev->addr_len);
	}

	return -MBTX_HEADER_LEN;	/* incomplete header, needs rebuild */
}

static int masc_dev_rebuild_header(void *hdrbuf, struct device *dev,
		unsigned long dest, struct sk_buff *skb)
{
	unsigned char *hdr = hdrbuf;

	MASC_DBG_PRINTK((KERN_INFO MNAME "masc_dev_rebuild_header called\n"));

	if (hdr[11] != MBTX_P_IP) {
		printk(KERN_ERR MNAME "%s cannot resolve HPID %d address\n",
				dev->name, hdr[11]);

		return 1;
	}

#ifdef CONFIG_INET
	if (arp_find(hdr + 3, dest, dev, dev->pa_addr, skb))
		return 1;
#endif
	return 0;
}

/*
 * ACK timeout
 */

static void masc_ack_timer(unsigned long arg)
{
	masc_dev_t *masc = (masc_dev_t *) arg;	/* ARGH! Ugly! Ugly! */

	if (masc == NULL)
		return;

	MASC_DBG_PRINTK((KERN_INFO MNAME "%s: ack timer\n", masc->if_name));

	del_timer(&masc->ack_timer);

	if (skb_queue_len(&masc->ackw_queue) > 0) {
		masc_send_ctrl(masc, "!");		/* send RACK */
		set_bit(MASC_FLG_RACKS, &masc->flags);	/* RACK sent flag	*/
		masc_kick_tty(masc);
		masc->ack_timer.expires = jiffies + MASC_ACK_TIMEOUT*HZ;
		add_timer(&masc->ack_timer);
	}
}

/*
 * Battery/RSSI timer
 */

static void masc_bat_timer(unsigned long arg)
{
	masc_dev_t *masc = (masc_dev_t *) arg;	/* Blech! */
	int do_kick = 0;

	if (masc == NULL)
		return;

	MASC_DBG_PRINTK((KERN_INFO MNAME "%s: battery/RSSI timer\n", masc->if_name));

	del_timer(&masc->bat_timer);

	if (clear_bit(MASC_FLG_HAVEBATT, &masc->flags)) {
		MASC_DBG_PRINTK((KERN_INFO MNAME "%s: sending battery status request\n", masc->if_name));
		masc_send_info(masc, masc_getbatt_frame, sizeof masc_getbatt_frame);
		do_kick = 1;
	}

	if (clear_bit(MASC_FLG_HAVERSSI, &masc->flags)) {
		MASC_DBG_PRINTK((KERN_INFO MNAME "%s: sending signal strength request\n", masc->if_name));
		masc_send_info(masc, masc_getsign_frame, sizeof masc_getsign_frame);
		do_kick = 1;
	}

	if (do_kick)
		masc_kick_tty(masc);
	else
		MASC_DBG_PRINTK((KERN_INFO MNAME "%s: nothing to do\n", masc->if_name));

	masc->bat_timer.expires = jiffies + MASC_BAT_POLL*HZ;
	add_timer(&masc->bat_timer);
}

static void masc_traf_timer(unsigned long arg)
{
	int traf_curr, do_kick = 0;
	masc_dev_t *masc = (masc_dev_t *) arg;	/* Blech! */

	del_timer(&masc->traf_timer);
	traf_curr = masc->traf_curr + 1;
	if (traf_curr >= sizeof masc->traf_hist / sizeof masc->traf_hist[0])
		traf_curr = 0;
	if ((masc->traf_count -= masc->traf_hist[traf_curr]) < MASC_MAX_AIRED_PACKETS)
		do_kick = 1;
	masc->traf_hist[traf_curr] = 0;
	masc->traf_curr = traf_curr;
	if (do_kick)
		masc_kick_tty(masc);
	masc->traf_timer.expires = jiffies + 1*HZ;
	add_timer(&masc->traf_timer);
}

/*
 * Initialize the ``base class'' portion of a masc device. This
 * is called after the masc object is registered as a network
 * device.
 */

static int masc_dev_init(struct device *dev)
{
	masc_dev_t *masc = dev->priv;
	int indx;
	int num = dev->base_addr;

	/* sanity check */

	if (!masc || masc->magic != MASC_MAGIC) {
		printk(KERN_ERR MNAME "net device isn't a masc device!");
		return -ENODEV;
	}

	dev->hard_header = masc_dev_hard_header;
	dev->rebuild_header = masc_dev_rebuild_header;
	dev->hard_header_len = MBTX_HEADER_LEN;

	dev->mtu = MASC_MTU;
	dev->hard_start_xmit = masc_dev_xmit;
	dev->open = masc_dev_open;
	dev->stop = masc_dev_stop;
	dev->get_stats = masc_dev_stats;
	dev->do_ioctl = masc_dev_ioctl;
	dev->addr_len = 3;
	dev->tx_queue_len = MASC_MAX_DEV_QUEUE;
	/* Pretend we are ethernet from an ARP point of view	*/
	/* Safe bet, but looks ridiculous!			*/
	dev->type = ARPHRD_ETHER;

	for (indx = 0; indx < DEV_NUMBUFFS; indx++)
		skb_queue_head_init(&dev->buffs[indx]);

	dev->flags = 0;
	if (masc_loopback)
		dev->flags |= IFF_LOOPBACK;
	dev->family = AF_INET;
	dev->pa_addr = 0;
	dev->pa_brdaddr = 0;
	dev->pa_mask = 0;
	dev->pa_alen = 4;

	/*
	 * Initialize the hardware broadcast device according
	 * to the the user-configurable values that can be
	 * specified on the insmod command line.
	 */

	if (num <= 9) {
		dev->broadcast[0] = masc_bcast[num] >> 16 & 0xff;
		dev->broadcast[1] = masc_bcast[num] >> 8 & 0xff;
		dev->broadcast[2] = masc_bcast[num] & 0xff;
	}

	return 0;

}

/*
 * Find an existing, free masc device on the list.  We don't bother locking the
 * list because it is updated by atomic operations, and nothing is ever deleted
 * from it until the module is removed from memory.  Thus the masc pointer can
 * never go invalid from under us, and masc->next is either a pointer to the
 * next node or a pointer to masc_list, and the access to the pointer object
 * is atomic.
 */

static masc_dev_t *masc_find_free(void)
{
	masc_dev_t *masc = masc_list.next;

	while (masc != &masc_list) {
		if (!set_bit(MASC_FLG_OCCUP, &masc->flags))
			return masc;
		masc = masc->next;
	}

	return NULL;
}


/*
 * Allocate masc device using kmalloc(), initialize it, register it as a
 * network device and add it to the global masc device list.
 */

static masc_dev_t *masc_dev_alloc(void)
{
	static struct device blank_dev = { 0 };
	static masc_dev_t blank_masc = { MASC_MAGIC };
	masc_dev_t *masc;
	int masc_num;

	blank_masc.flags |= (1 << MASC_FLG_OCCUP);

	if ((masc = masc_find_free()) != NULL) {
		masc_num = masc->line;
		blank_masc.next = masc->next;
		blank_masc.prev = masc->prev;
		*masc = blank_masc;
	} else {
		masc = kmalloc(sizeof *masc, GFP_KERNEL);
		if (!masc) {
			printk(KERN_ERR MNAME "failure to allocate masc_dev!\n");
			goto failed_masc;
		}
		masc_num = masc_numdev++;
		*masc = blank_masc;
		masc->next = &masc_list;
		masc->prev = masc_list.prev;
		masc_list.prev->next = masc;
		masc_list.prev = masc;
	}

	if (!(masc->tx_buf = kmalloc(MASC_BUF_SIZE, GFP_KERNEL))) {
		printk(KERN_ERR MNAME "failure to allocate tx buffer\n");
		goto failed_tx_buff;
	}

	if (!(masc->rx_buf = kmalloc(MASC_BUF_SIZE, GFP_KERNEL))) {
		printk(KERN_ERR MNAME "failure to allocate rx buffer\n");
		goto failed_rx_buff;
	}

	sprintf(masc->if_name, "masc%d", masc_num);

	if (dev_get(masc->if_name) != NULL) {
		printk(KERN_ERR MNAME "name %s is already in use!\n", masc->if_name);
		goto failed_name_grab;
	}

	masc->tx_head = masc->tx_buf;
	masc->line = masc_num;
	masc->send_ack = -1;
	masc->last_ack = -1;
	init_timer(&masc->ack_timer);
	masc->ack_timer.data = (unsigned long) masc;
	/* ARGH! The above cast is so lame, I want to scream!!! */
	masc->ack_timer.function = masc_ack_timer;	
	init_timer(&masc->bat_timer);
	masc->bat_timer.data = (unsigned long) masc;
	masc->bat_timer.function = masc_bat_timer;	
	masc->traf_timer.data = (unsigned long) masc;
	masc->traf_timer.function = masc_traf_timer;	
	masc->dev_base = blank_dev;
	masc->dev_base.name = masc->if_name;
	masc->dev_base.base_addr = masc_num;
	masc->dev_base.priv = masc;	
	masc->dev_base.next = NULL;
	masc->dev_base.init = masc_dev_init;
	masc->dev = &masc->dev_base;
	skb_queue_head_init(&masc->mpak_queue);
        skb_queue_head_init(&masc->info_queue);
        skb_queue_head_init(&masc->ctrl_queue);
        skb_queue_head_init(&masc->free_queue);
        skb_queue_head_init(&masc->ackw_queue);
        skb_queue_head_init(&masc->fh_queue);
	if (register_netdev(masc->dev) != 0) {
		printk(KERN_ERR MNAME "failed to register network device!\n");
		goto failed_register;
	}
	return masc;

failed_register:
	kfree(masc->rx_buf);
failed_name_grab:
failed_rx_buff:
	kfree(masc->tx_buf);
failed_tx_buff:
	/* must not free masc here, it's been put on the list */
failed_masc:
	return NULL;
}

/*
 * Deallocate a masc structure
 */

static void masc_dev_dealloc(masc_dev_t *masc)
{
	struct tty_struct *tty = masc->tty;

	/*
	 * Break the tty --> masc link
	 */

	if (tty && tty->disc_data == masc)
		tty->disc_data = NULL;


	masc->tty = NULL;

	/*
	 * Kill timers!
	 */

	del_timer(&masc->ack_timer);
	del_timer(&masc->bat_timer);
	del_timer(&masc->traf_timer);

	/*
	 * Close net device, and unregister it
	 */

	dev_close(masc->dev);
	unregister_netdev(masc->dev);

	/*
	 * Free memory resources---all except for the masc_dev_t structure
	 * itself. This stays on the global list, and is reused.
	 */

	masc_free_buffers(masc);
	kfree(masc->rx_buf);
	kfree(masc->tx_buf);

	/*
	 * Setting the flags to zero marks this structure
	 * as free, so that the next masc_find_free()
	 * might be able to grab it.
	 */

	masc->flags = 0;
}

/*
 * The masc_tty_* functions form an implementation of the
 * abstract line discipline interface that is registered with the TTY
 * subsystem.
 */

/*
 * Called when the line discipline of a tty becomes MASC.
 */

static int masc_tty_open(struct tty_struct *tty)
{
	masc_dev_t *masc;

	if (tty->disc_data) {
		return -EEXIST;
	}

	if (!(masc = masc_dev_alloc())) {
		return -ENFILE;
	}

	masc->tty = tty;	
	tty->disc_data = masc;

	if (tty->ldisc.flush_buffer)
		tty->ldisc.flush_buffer (tty);

	if (tty->driver.flush_buffer)
		tty->driver.flush_buffer (tty);

	/*
	 * When the radio modem gives us its terminal MAN 
	 * address, we will start the device.
	 */

	masc->dev->tbusy = 1;

	MASC_DBG_PRINTK((KERN_INFO MNAME "allocated %s\n", masc->if_name));

	/*
	 * Send INIT frame to the radio modem
	 */

	masc_send_info(masc, masc_init_frame, sizeof masc_init_frame);
	masc_kick_tty(masc);

	/*
	 * Start battery and signal strength interrogation timer.
	 * The first time around, we must pretend that we have
	 * this info already, so we set the appropriate flags.
	 * When the timer goes off, it clears the flags and sends
	 * the request if the flags were previously set. If they
	 * are already clear, it does not duplicate the request,
	 * (to prevent spamming the queue if the modem is not there).
	 */

	set_bit(MASC_FLG_HAVERSSI, &masc->flags);
	set_bit(MASC_FLG_HAVEBATT, &masc->flags);

	/*
	 * For the first sample, we set the timer for just
	 * one second, so the info is available readily.
	 */

	masc->bat_timer.expires = jiffies + 1*HZ;
	add_timer(&masc->bat_timer);

	/*
	 * Traffic histogram timer is once every second over
	 * lifetime of driver.
	 */

	masc->traf_timer.expires = jiffies + 1*HZ;
	add_timer(&masc->traf_timer);

	MOD_INC_USE_COUNT;

	return masc->line;
}

/*
 * Called when the MASC line discipline is uninstalled
 * from the tty (because for instance the tty has closed).
 */

static void masc_tty_close(struct tty_struct *tty)
{
	masc_dev_t *masc = tty->disc_data;

	MASC_DBG_PRINTK((KERN_INFO MNAME "masc_tty_close called for %s\n", masc->if_name));

	if (!masc) {
		printk(KERN_ERR MNAME "tty struct disc_data pointer is null\n");
		return;
	}

	masc_dev_dealloc(masc);

	MASC_DBG_PRINTK((KERN_INFO MNAME "deallocated %s\n", masc->if_name));
	MOD_DEC_USE_COUNT;
}

/*
 * We don't need to provide these functions unless we allow a process to
 * read/write the tty concurrently with the networking system doing the same.
 * Since we provide full MASC functionality, we don't need to do this.
 * PPP does provide these functions because the complex authentication
 * protocols are handled by the PPP daemon.
 */

#if 0
static int masc_tty_read(struct tty_struct *tty, struct file *file,
		unsigned char *buf, unsigned int nr)
{
}

static int masc_tty_write(struct tty_struct *tty, struct file *file,
		const unsigned char *buf, unsigned int nr)
{
}

static int masc_tty_ioctl(struct tty_struct *tty, struct file *file,
		unsigned int cmd, unsigned long arg)
{
}


static int masc_tty_select(struct tty_struct *tty, struct inode *inode,
		struct file *file, int sel_type, struct select_table_struct *wait)
{
}

#endif

/*
 * This function is called from the tty system to pass data to us. We
 * are given a count of bytes, and a pointer to the data (cp). Also, there
 * is a second pointer (fp) which points to an array of bytes which indicate
 * indicates which bytes have framing (parity) errors.
 * We assume that whenever we see a carriage return, we have received
 * a MASC frame, so we call the frame processing function. When this
 * happens, the receive buffer is cleared and we start again.
 * This isn't right, strictly speaking, since we should scan for
 * the leading ^ which kicks off a MASC frame.
 */

static void masc_tty_receive_buf(struct tty_struct *tty, const unsigned char *cp,
		char *fp, int count)
{
	masc_dev_t *masc = tty->disc_data;

	while (count--) {
		unsigned int ch = *cp++;
		if (fp && *fp++) {
			if (!set_bit(MASC_FLG_TOSS, &masc->flags)) {
				masc->rx_errors++;
			}
			continue;
		}

		if (masc->rx_count < MASC_BUF_SIZE) {
			masc->rx_buf[masc->rx_count++] = ch;
		} else {
			if (!set_bit(MASC_FLG_OFLOW, &masc->flags)) {
				masc->rx_errors++;
				printk(KERN_ERR MNAME
				"receive buffer overflow (%s)---truncating and tossing packet\n", masc->if_name);
			}
		}

		if (ch == '\r')
			masc_process_frame(masc);
	}
}

/*
 * The tty system uses this function to query how much buffer space we have. We
 * simply say that we can take a lot. This is true, since we remove and stuff
 * it into a dynamically allocated packets that go to the network layer.
 */

static int masc_tty_receive_room(struct tty_struct *tty)
{
	return 65536;
}

/*
 * Transmit completion handler, called by the tty subsystem to indicate that it
 * has completed a write operation. Our response is to stuff more data into the
 * tty and call the tty's write method to continue the flow. If no data is
 * available, as indicated by masc_fill_transmit_buf() returning zero, we lower
 * the device busy transmitting flag, and set a flag for the bottom half
 * processing to tell the network layer to try sending us more stuff.
 */

static void masc_tty_write_wakeup(struct tty_struct *tty)
{
	masc_dev_t *masc = tty->disc_data;
	int nwrit;

	if (!masc || !masc->magic == MASC_MAGIC)
		return;

	MASC_DBG_PRINTK((KERN_INFO MNAME "%s: write wakeup\n", masc->if_name));

	if (!masc_fill_transmit_buf(masc)) {
		MASC_DBG_PRINTK((KERN_INFO MNAME "%s: end of write burst\n", masc->if_name));
		tty->flags &= ~(1 << TTY_DO_WRITE_WAKEUP);
		clear_bit(MASC_FLG_TBUSY, &masc->flags);
		return;
	}

	masc->dev->trans_start = jiffies;
	nwrit = tty->driver.write(tty, 0, masc->tx_head, masc->tx_left);
	masc->tx_left -= nwrit;
	masc->tx_head += nwrit;
}

/*
 * Generate the /proc/net/mobitex_masc data.
 */

int masc_get_info(char *buffer, char **start, off_t offset, int length, int dummy)
{
        int len = 0;
	masc_dev_t *masc;

        len += sprintf(buffer, "Name     Batt Sign Roam Stop LastFourErr LocalMAN BcastMAN Notr Illg Cong Errs\n");

	for (masc = masc_list.next; masc != &masc_list; masc = masc->next) {
		struct device *dev = masc->dev;

		if (!test_bit(MASC_FLG_OCCUP, &masc->flags))
			continue;

		/*
		 * cli()/sti() is needed here to prevent the masc driver
		 * code from being called from an interrupt and messing with
		 * these values, so we get a consistent snapshot.
		 */

		cli();
		len += sprintf(buffer + len, "%-8s %4d %4d %4s %4s %02X,%02X,%02X,%02X "
			"%02X:%02X:%02X %02X:%02X:%02X %4lu %4lu %4lu %4lu\n",
			masc->if_name, masc->batt, masc->sign,
			test_bit(MASC_FLG_NOCONTACT, &masc->flags) ? "Yes" : "No",
			test_bit(MASC_FLG_TSTOP, &masc->flags) ? "Yes" : "No",
			(unsigned int) masc->errhist[0], (unsigned int) masc->errhist[1],
			(unsigned int) masc->errhist[2], (unsigned int) masc->errhist[3],
			(unsigned int) dev->dev_addr[0], (unsigned int) dev->dev_addr[1],
			(unsigned int) dev->dev_addr[2],
			(unsigned int) dev->broadcast[0], (unsigned int) dev->broadcast[1],
			(unsigned int) dev->broadcast[2],
			masc->rx_notran, masc->rx_illegal, masc->rx_congest, masc->rx_errcount);
		sti();

		/*
		 * the buffer is supposedly only a page long, so if we get
		 * close to the (i386) page size, we bail. This is unlikely,
		 * since it would require like 50 masc devices.
		 */

		if (len > 4000)
			break;
	}

	*start = buffer + offset; 
	len -= offset;

	if (len < 0)
		len = 0;

	if (len>length)
		len = length; 

	return len;
}

/*
 * Initialize the MASC subsystem, and register it as a TTY line discipline
 * module. Subsequent opening and closing of MASC devices is initiated from the
 * TTY subsystem, when a TTY is instructed to register MASC as a line
 * discipline.  When a TTY is told to install MASC line discipline, it opens a
 * MASC device, A MASC device instance is allocated and registers itself as a
 * network driver. This creates a bridge between the TTY and the kernel's
 * networking subsystem, with MASC device serving as an intermediary. Neat, huh?
 */

static int masc_do_init(void)
{
	int status = 0;	

	printk(KERN_INFO MNAME "version %s\n", MASC_VERSION);

	if (masc_maxdev < 4) {
		printk(KERN_INFO MNAME "masc_maxdev too small!\n");
		printk(KERN_INFO MNAME "was %d\n", masc_maxdev);
		masc_maxdev = 4;
		printk(KERN_INFO MNAME "set to %d\n", masc_maxdev);
	}

#ifdef CONFIG_PROC_FS
	masc_proc.namelen = sizeof "mobitex_masc" - 1;
	masc_proc.name = "mobitex_masc";
	masc_proc.mode = S_IFREG | S_IRUGO;
	masc_proc.nlink = 1;
	masc_proc.uid = 0;
	masc_proc.gid = 0;
	masc_proc.size = 0;
	masc_proc.ops = &proc_net_inode_operations;
	masc_proc.get_info = masc_get_info;
	masc_proc.fill_inode = NULL;
	masc_proc.next = NULL;
	masc_proc.parent = NULL;
	masc_proc.subdir = NULL;
	masc_proc.data = NULL;
	masc_proc_flag = proc_register_dynamic(&proc_net, &masc_proc);
	if (masc_proc_flag < 0) {
		printk(KERN_INFO MNAME "couldn't register /proc entry\n");
	}
#endif

	masc_list.next = &masc_list;
	masc_list.prev = &masc_list;

	masc_ldisc.magic = TTY_LDISC_MAGIC;
	masc_ldisc.open	= masc_tty_open;
	masc_ldisc.close = masc_tty_close;
	masc_ldisc.flush_buffer = NULL;	
	masc_ldisc.chars_in_buffer = NULL;
	masc_ldisc.read	= NULL;	
	masc_ldisc.write= NULL;
	masc_ldisc.ioctl= NULL;
	masc_ldisc.select = NULL; 
	masc_ldisc.set_termios = NULL;
	masc_ldisc.receive_room = masc_tty_receive_room;
	masc_ldisc.receive_buf = masc_tty_receive_buf;
	masc_ldisc.write_wakeup = masc_tty_write_wakeup;

	if ((status = tty_register_ldisc(N_MASC, &masc_ldisc)) == 0) {
		printk(KERN_INFO MNAME "registered line discipline\n");
	} else {
		printk(KERN_ERR MNAME "cannot register line discipline (error = %d)\n", status);
		return status;
	}

	return status;
}


/*
 * If this is built as a module, the init_module() and cleanup_module()
 * functions are used as entry points to initialize and deinitialize the MASC
 * subsystem.
 *
 * On the other hand, if this is statically linked to the kernel, the single
 * function masc_init() is provided for the kernel to call at system
 * initialization time. A static driver persists until the operating system is
 * terminated, so a matching cleanup function is not needed
 */

#ifdef MODULE

int init_module(void)
{
	return masc_do_init();
}

void cleanup_module(void)
{
	int status = 0;
	masc_dev_t *masc, *next;

	if (tty_register_ldisc(N_MASC, NULL) == 0) {
		printk(KERN_INFO
		       MNAME "line discipline successfully unregistered\n");
	} else {
		printk(KERN_ERR
			MNAME "unable to unregister MASC line discipline "
			"(error = %d)\n", status);
	}

#ifdef CONFIG_PROC_FS
	if (masc_proc_flag == 0)
		proc_unregister(&proc_net, masc_proc.low_ino);
#endif

	for (masc = masc_list.next; masc != &masc_list; masc = next) {
		next = masc->next;
		kfree(masc);
	}

	printk(KERN_INFO MNAME "module cleaned up\n");
}

#else	/* if !defined MODULE */

int masc_init(struct device *not_used)
{
	int status = masc_do_init();
	return (status == 0) ? ENODEV : status;
}

#endif	/* !defined MODULE */
